"use strict";
var Promise = require('bluebird');
var intersection = require('lodash/intersection');
var assign = require('lodash/assign');
function buildSynchronizer(spec, options) {
    var pool = null;
    function synchronizer(transaction, item, errors) {
        if (item.action === 'INSERT') {
            return insert(pool, spec.name, options && options.insertErrorMessageBuilder, transaction, item, errors);
        }
        if (item.action === 'UPDATE') {
            return update(pool, spec.name, options && options.updateErrorMessageBuilder, transaction, item, errors);
        }
        if (item.action === 'REMOVE') {
            return remove(pool, spec.name, options && options.removeErrorMessageBuilder, transaction, item, errors);
        }
        return Promise.reject(new Error('provided item is invalid'));
    }
    function initialize(p) {
        pool = p;
    }
    return { synchronizer: synchronizer, initialize: initialize };
}
exports.buildSynchronizer = buildSynchronizer;
function insert(pool, model, messageBuilder, transaction, item, errors) {
    var promise = pool.models[model].sequelizeModel.create(item.data, {
        transaction: transaction,
        fields: pool.models[model].allowedInsertAttributes
    }).then(function (instance) {
        var tasks = [];
        var aliasModelMap = pool.models[model].aliasModelMap;
        var aliasKeyMap = pool.models[model].aliasKeyMap;
        Object.keys(aliasModelMap).forEach(function (alias) {
            var nested = item.data[alias];
            if (nested) {
                if (Array.isArray(nested)) {
                    nested.forEach(function (nestedItem) {
                        tasks.push(pool.models[aliasModelMap[alias]].synchronizer(transaction, {
                            action: 'INSERT',
                            itemId: item.itemId,
                            model: aliasModelMap[alias],
                            data: assign({}, nestedItem, (_a = {}, _a[aliasKeyMap[alias]] = instance[pool.models[model].primaryKey], _a))
                        }, errors));
                        var _a;
                    });
                }
                else {
                    tasks.push(pool.models[aliasModelMap[alias]].synchronizer(transaction, {
                        action: 'INSERT',
                        itemId: item.itemId,
                        model: aliasModelMap[alias],
                        data: assign({}, nested, (_a = {}, _a[aliasKeyMap[alias]] = instance[pool.models[model].primaryKey], _a))
                    }, errors));
                }
            }
            var _a;
        });
        return tasks.length > 0 ? Promise.all(tasks) : undefined;
    });
    if (typeof messageBuilder === 'function') {
        return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
    }
    return promise;
}
function update(pool, model, messageBuilder, transaction, item, errors) {
    var primaryKey = pool.models[model].primaryKey;
    var promise = pool.models[model].sequelizeModel.update(item.data, {
        transaction: transaction,
        validate: true,
        fields: buildUpdateAttributes(pool.models[model].allowedUpdateAttributes, item.attributes),
        where: (_a = {}, _a[primaryKey] = item.data[primaryKey], _a)
    });
    if (typeof messageBuilder === 'function') {
        return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
    }
    return promise;
    var _a;
}
function remove(pool, model, messageBuilder, transaction, item, errors) {
    var promise = pool.models[model].sequelizeModel.destroy({
        transaction: transaction,
        where: { id: item.targetId }
    });
    if (typeof messageBuilder === 'function') {
        return promise.catch(function (error) { return errors.push({ itemId: item.itemId, message: messageBuilder(item, error) }); });
    }
    return promise;
}
function buildUpdateAttributes(allowedAttributes, updateAttributes) {
    if (Array.isArray(allowedAttributes)) {
        return intersection(allowedAttributes, updateAttributes);
    }
    return updateAttributes;
}
